"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var schema_1 = require("./schema");
var clock_1 = require("./clock");
var system_1 = require("../scheduler/system");
var DEFAULT_TICK_RATE = 60;
var DEFAULT_PING_RATE = 300;
var DEFAULT_PING_BUFFER_SIZE = 1024;
var DEFAULT_FRAME_SKIP = Infinity;
var scratchBuffer = [];
function compareNum(a, b) { return a - b; }
function calcDelay(samples) {
    var N = scratchBuffer.length = samples.length;
    var m0 = 0;
    var m1 = 0;
    for (var i = 0; i < N; ++i) {
        var x = scratchBuffer[i] = samples[i];
        m0 += x;
        m1 += x * x;
    }
    m0 /= N;
    m1 /= N;
    var sigma = Math.sqrt(m1 - m0 * m0);
    scratchBuffer.sort(compareNum);
    var med = scratchBuffer[N >>> 1];
    var total = 0;
    var count = 0;
    for (var i = 0; i < samples.length; ++i) {
        var x = samples[i];
        if (Math.abs(x - med) <= sigma) {
            total += x;
            count++;
        }
    }
    return total / count;
}
var MuClockClient = (function () {
    function MuClockClient(spec) {
        var _this = this;
        this._clock = new clock_1.MuClock();
        this._delay = 0;
        this._pingSamples = [];
        this._numPings = 0;
        this.tickRate = DEFAULT_TICK_RATE;
        this.ping = 0;
        this.skippedFrames = 0;
        this.frameSkip = DEFAULT_FRAME_SKIP;
        this.monotone = true;
        this._started = false;
        this._tickCount = 0;
        this._poll = function () {
            _this._pollHandle = _this.scheduler.requestIdleCallback(_this._poll, { timeout: 0.5 * _this.tickRate });
            _this.poll();
        };
        this._lastNow = 0;
        this._protocol = spec.client.protocol(schema_1.MuClockProtocol);
        this._onTick = spec.tick || function () { };
        if (spec.frameSkip) {
            this.frameSkip = spec.frameSkip | 0;
        }
        this.monotone = !spec.enableRewind;
        this.maxPingSamples = spec.pingBufferSize || DEFAULT_PING_BUFFER_SIZE;
        this.scheduler = spec.scheduler || system_1.MuSystemScheduler;
        this._protocol.configure({
            message: {
                init: function (_a) {
                    var tickRate = _a.tickRate, serverClock = _a.serverClock, skippedFrames = _a.skippedFrames;
                    _this._clock.reset(serverClock);
                    _this.tickRate = tickRate;
                    _this.skippedFrames = skippedFrames;
                    _this._tickCount = Math.floor(serverClock / tickRate) - skippedFrames;
                    _this._started = true;
                    for (var i = 0; i < 5; ++i) {
                        _this._doPing();
                    }
                    _this._pingInterval = _this.scheduler.setInterval(function () { return _this._doPing(); }, spec.pingRate || DEFAULT_PING_RATE);
                    _this._pollHandle = _this.scheduler.requestIdleCallback(_this._poll, { timeout: 0.5 * _this.tickRate });
                    if (spec.ready) {
                        spec.ready();
                    }
                },
                frameSkip: function (skippedFrames) {
                    _this.skippedFrames = Math.max(skippedFrames, _this.skippedFrames);
                },
                pong: function (_a) {
                    var clientClock = _a.clientClock, serverClock = _a.serverClock, skippedFrames = _a.skippedFrames;
                    var now = _this._clock.now();
                    var delay = serverClock - 0.5 * (now + clientClock);
                    _this.ping = now - clientClock;
                    _this.skippedFrames = Math.max(_this.skippedFrames, skippedFrames);
                    if (_this._pingSamples.length < _this.maxPingSamples) {
                        _this._pingSamples.push(delay);
                        _this._numPings++;
                    }
                    else {
                        _this._pingSamples[_this._numPings++ % _this._pingSamples.length] = delay;
                    }
                    _this._delay = calcDelay(_this._pingSamples);
                },
            },
            close: function () {
                _this.scheduler.cancelIdleCallback(_this._pollHandle);
                _this.scheduler.clearInterval(_this._pingInterval);
            },
        });
    }
    MuClockClient.prototype._doPing = function () {
        this._protocol.server.message.ping(this._clock.now());
    };
    MuClockClient.prototype._simulationClock = function () {
        var remoteClock = Math.max(this._clock.now() + this._delay, this._lastNow + 1e-6);
        this._lastNow = remoteClock;
        return remoteClock;
    };
    MuClockClient.prototype.poll = function () {
        if (!this._started) {
            return;
        }
        var ticksSmooth = this._simulationClock() / this.tickRate - this.skippedFrames;
        var targetTickCount = Math.floor(ticksSmooth);
        if (targetTickCount > 0) {
            var ticks = targetTickCount - this._tickCount;
            if ((ticks > this.frameSkip) ||
                (ticks < 0 && !this.monotone && targetTickCount > 0)) {
                this._tickCount = targetTickCount;
                this._onTick(this._tickCount);
            }
            else {
                while (this._tickCount < targetTickCount) {
                    this._onTick(++this._tickCount);
                }
            }
        }
    };
    MuClockClient.prototype.now = function () {
        if (!this._started) {
            return 0;
        }
        return this._clock.now();
    };
    MuClockClient.prototype.tick = function () {
        if (!this._started) {
            return 0;
        }
        return Math.max(0, Math.min(this._tickCount + 1, this._simulationClock() / this.tickRate - this.skippedFrames));
    };
    return MuClockClient;
}());
exports.MuClockClient = MuClockClient;
//# sourceMappingURL=client.js.map