"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var clock_1 = require("./clock");
var schema_1 = require("./schema");
var system_1 = require("../scheduler/system");
var DEFAULT_TICK_RATE = 60;
var DEFAULT_FRAMESKIP = Infinity;
var MuClockServer = (function () {
    function MuClockServer(spec) {
        var _this = this;
        this._clock = new clock_1.MuClock();
        this._tickCount = 0;
        this.frameSkip = DEFAULT_FRAMESKIP;
        this.skippedFrames = 0;
        this._poll = function () {
            _this._pollHandle = _this.scheduler.requestIdleCallback(_this._poll, { timeout: 0.5 * _this.tickRate });
            _this.poll();
        };
        this.tickRate = spec.tickRate || DEFAULT_TICK_RATE;
        this._protocol = spec.server.protocol(schema_1.MuClockProtocol);
        this._onTick = spec.tick || function () { };
        if ('frameSkip' in spec) {
            this.frameSkip = +(spec.frameSkip || 0);
        }
        this.scheduler = spec.scheduler || system_1.MuSystemScheduler;
        this._protocol.configure({
            ready: function () {
                _this._clock.reset(0);
                _this._pollHandle = _this.scheduler.requestIdleCallback(_this._poll, { timeout: 0.5 * _this.tickRate });
            },
            message: {
                ping: function (client, clientClock) {
                    var msg = schema_1.MuPingResponseSchema.alloc();
                    msg.clientClock = clientClock;
                    msg.serverClock = _this._clock.now();
                    msg.skippedFrames = _this.skippedFrames;
                    client.message.pong(msg);
                    schema_1.MuPingResponseSchema.free(msg);
                },
            },
            connect: function (client) {
                client.message.init({
                    tickRate: _this.tickRate,
                    serverClock: _this._clock.now(),
                    skippedFrames: _this.skippedFrames,
                });
            },
            close: function () {
                _this.scheduler.cancelIdleCallback(_this._pollHandle);
            },
        });
    }
    MuClockServer.prototype.poll = function () {
        var now = this._clock.now();
        var ticksSmooth = now / this.tickRate - this.skippedFrames;
        var targetTick = Math.floor(ticksSmooth);
        var numTicks = targetTick - this._tickCount;
        if (numTicks <= this.frameSkip) {
            while (this._tickCount < targetTick) {
                this._onTick(++this._tickCount);
            }
        }
        else if (numTicks > 0) {
            this.skippedFrames += targetTick - this._tickCount - 1;
            this._tickCount = Math.floor(now / this.tickRate - this.skippedFrames);
            this._protocol.broadcast.frameSkip(this.skippedFrames, true);
            this._onTick(this._tickCount);
        }
    };
    MuClockServer.prototype.tick = function () {
        if (this._clock) {
            var t = this._clock.now() / this.tickRate - this.skippedFrames;
            return Math.min(t, this._tickCount + 1);
        }
        return 0;
    };
    MuClockServer.prototype.now = function () {
        if (this._clock) {
            return this._clock.now();
        }
        return 0;
    };
    return MuClockServer;
}());
exports.MuClockServer = MuClockServer;
//# sourceMappingURL=server.js.map